      Program Appendix_7_2_1
C
C  This is a main program for the example 4. The example computes
C  the following equations in parallel:
C
C     [C]=[A][B]
C     {Y}=[C]{X}
C
C  where matrices [A] and [B] are of order (NxN), and vector {X} is known.
C  The example uses a parallel lock to handle the critical sections
C  in both equations. Task synchronization is handled by
C  a WaitFor routine and a parallel event.
C
C  global variables
C
      INTEGER*4 Order
      PARAMETER (Order=220)
      INTEGER*4 NumberOfTasks
      PARAMETER (NumberOfTasks=6)
      INTEGER*4 GrainSize
      PARAMETER (GrainSize=2500)
      INTEGER*4 N,Outstanding,LockID,EventID,SubsetA,SubsetB
      INTEGER*4 J4TEMP,Seed,TaskID(NumberOfTasks)
      REAL*4 A(Order,Order),B(Order,Order),C(Order,Order)
      REAL*4 X(Order),Y(Order)
C
C  declare dispatched subroutine
C
      EXTERNAL SubA
C
C  generate data in a sequential procedure
C
      WRITE(*,'('' >>> Generate data ...'')')
      Seed=111111
      N=Order
      DO J4TEMP=1,Order*Order
         A(J4TEMP,1)=Ran(Seed)
      END DO
      DO J4TEMP=1,Order*Order
         B(J4TEMP,1)=Ran(Seed)
      END DO
      DO J4TEMP=1,Order
         X(J4TEMP)=Ran(Seed)
      END DO
C
C  create parallel lock
C
      CALL CreateParallelLock(LockID)
C
C  create parallel event
C
      CALL CreateParallelEvent(EventID)
C
C  determine length of subset
C  (where "N*(2..." please see subroutine SubA)
C
      SubsetA=MIN0(GrainSize/(N*(2*N+1)),Order/NumberOfTasks)
      IF(SubsetA.EQ.0) SubsetA=1
      SubsetB=MIN0(GrainSize/(2*N+1),Order/NumberOfTasks)
      IF(SubsetB.EQ.0) SubsetB=1
C
C  initialize
C
      Outstanding=1
      CALL InitializeParallelEvent(EventID,NumberOfTasks)
C
C  compute in parallel
C
      WRITE(*,'('' >>> Compute in parallal ...'')')
      DO J4TEMP=1,NumberOfTasks-1
         CALL Dispatch_11(TaskID(J4TEMP),SubA,A,
     &                                        B,
     &                                        C,
     &                                        X,
     &                                        Y,
     &                                        N,
     &                                        Outstanding,
     &                                        LockID,
     &                                        EventID,
     &                                        SubsetA,
     &                                        SubsetB)
      END DO
      CALL SubA(A,B,C,X,Y,N,Outstanding,LockID,EventID,
     &          SubsetA,SubsetB)
C
C  wait for the completion of tasks
C
      CALL WaitForAndTerminateAllTasks
C
C  delete parallel lock
C
      CALL DeleteParallelLock(LockID)
C
C  delete parallel event
C
      CALL DeleteParallelEvent(EventID)
C
C  output
C
      WRITE(*,'(''     Y(  1)='',E12.6)') Y(1)
      WRITE(*,'(''     Y(100)='',E12.6)') Y(100)
      WRITE(*,'(''     Y(200)='',E12.6)') Y(200)
C
      STOP
      END
      SUBROUTINE SubA(A,B,C,X,Y,N,Outstanding,LockID,EventID,
     &                SubsetA,SubsetB)
C
C
C  routine to compute vector {Y} in parallel
C  (A)FORTRAN CALL: CALL SubA(A,B,C,X,Y,N,Outstanding,LockID,EventID,
C                             SubsetA,SubsetB)
C     1.A,B: <R4> known matrices, dimension(N,N)
C     2.C: <R4> matrix to compute, dimension(N,N)
C     3.X: <R4> known vector, dimension(N)
C     4.Y: <R4> vector to compute, dimension(N)
C     5.N: <I4> order
C     6.Outstanding: <I4> outstanding element that must be initialized
C                         to one
C     7.LockID: <I4> lock identifier
C     8.EventID: <I4> identifier of parallel event
C     9.SubsetA: <I4> length of subset of [C]
C    10.SubsetB: <I4> length of subset of {Y}
C
C  dummy arguments
C
      INTEGER*4 N,Outstanding,LockID,EventID,SubsetA,SubsetB
      REAL*4 A(N,N),B(N,N),C(N,N),X(N),Y(N)
C
C  private variables
C
      INTEGER*4 J4TEMP$,K4TEMP$,L4TEMP$,M4TEMP$
      LOGICAL*4 YES$
      AUTOMATIC J4TEMP$,K4TEMP$,L4TEMP$,M4TEMP$,YES$
C
C  [critical section] -- the first equation
C  dynamic distribution
C
10    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 20
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+SubsetA
      CALL UnlockParallelLock(LockID)
C
C  [parallelism] -- the first equation
C  compute a partial rows of [C]
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+SubsetA-1,N)
         DO L4TEMP$=1,N
            C(L4TEMP$,K4TEMP$)=0.0
            DO M4TEMP$=1,N
               C(L4TEMP$,K4TEMP$)=C(L4TEMP$,K4TEMP$)+
     &                            A(L4TEMP$,M4TEMP$)*
     &                            B(M4TEMP$,K4TEMP$)
            END DO
         END DO
      END DO
C  where the grain size = SubsetA*N*(2N+1)
      GO TO 10
C
C  [task synchronization] -- the first equation
C  complete a cycle of parallel event
C
20    CALL PostParallelEvent(EventID)
      CALL IfCompleteParallelEvent(EventID,Yes$)
C
C  [subsequently sequential fragment] -- the first equation
C
      IF(.NOT.YES$) THEN
         Outstanding=1
         CALL CompleteParallelEvent(EventID)
      END IF
C
C  [critical section] -- the second equation
C  dynamic distribution
C
30    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         RETURN
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+SubsetB
      CALL UnlockParallelLock(LockID)
C
C  [parallelism]  -- the second equation
C  compute a partial elements of vector {Y}
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+SubsetB-1,N)
         Y(K4TEMP$)=0.0
         DO L4TEMP$=1,N
            Y(K4TEMP$)=Y(K4TEMP$)+C(K4TEMP$,L4TEMP$)*X(L4TEMP$)
         END DO
      END DO
C  where the grain size = SubsetB*(2N+1)
      GO TO 30
C
      END
      REAL*4 FUNCTION RAN(SEED)
C
C
C  function to generate a pseudo-random number
C  (A)FORTRAN ENTRY: RAN(SEED)
C     1.RAN: <R4> return a random number
C     2.SEED: <I4> a seed (updated)
C
      INTEGER*4 SEED,A
      REAL*8 RTEMPA,RTEMPB
      DATA M/2147483647/,A/16807/
C
C  generate a random number
C
      RTEMPA=DFLOAT(SEED)*A
      ITEMP=INT(RTEMPA/DFLOAT(M))
      RTEMPB=DFLOAT(M)*DFLOAT(ITEMP)
      SEED=INT(RTEMPA-RTEMPB)
      RAN=DFLOAT(SEED)/DFLOAT(M)
C
      RETURN
      END
 
