      Program Appendix_8_2_2
C
C
C  This is a main program for the example in section 8.2.2.
C  The example demonstrates a heterogeneous multitasking to compute
c  results: S1, S2, S3, and S4. There are 5 tasks originated:
C     Task 1: generate and send data to Tasks 2, 3, 4, and 5.
C     Task 2: compute S1, and send result to task 4.
C     Task 3: compute S2, and send result to task 4.
C     Task 4: compute S3, and send result to task 4.
C     Task 5: compute S4
C
C  global variables
C
      INTEGER*4 LengthOfBuffer
      PARAMETER (LengthOfBuffer=256)
      INTEGER*4 DataLength
      PARAMETER (DataLength=200000)
      INTEGER*4 PlateOne,PlateTwo,PlateThree,PlateFour
      INTEGER*4 SourceOne(2),SourceTwo(2),SourceThree(2),SourceFour(4)
      INTEGER*4 TaskOne,TaskTwo,TaskThree,TaskFour,TaskFive,K4TEMP
      INTEGER*4 BufferLength,LengthOfData,J4TEMP
      REAL*4 DataAreaOne(LengthOfBuffer,2),DataAreaTwo,DataAreaThree
      REAL*4 DataAreaFour,S1,S2,S3,S4
C
C  declare dispatched subroutines
C
      EXTERNAL Sub0,SubS1,SubS2,SubS3,SubS4
C
C  initialize
C
      BufferLength=LengthOfBuffer
      LengthOfData=DataLength
C
C  create data plates
C
      CALL CreateDataPlate(PlateOne,INT4(3))
      CALL CreateDataPlate(PlateTwo,INT4(1))
      CALL CreateDataPlate(PlateThree,INT4(1))
      CALL CreateDataPlate(PlateFour,INT4(1))
C
C  define "source"
C
      SourceOne(1)=PlateOne
      SourceOne(2)=0
      SourceTwo(1)=PlateOne
      SourceTwo(2)=0
      SourceThree(1)=PlateOne
      SourceThree(2)=0
      SourceFour(1)=PlateTwo
      SourceFour(2)=PlateThree
      SourceFour(3)=PlateFour
      SourceFour(4)=0
C
C  function partitioning
C
      CALL Dispatch_4(TaskOne,Sub0,DataAreaOne,BufferLength,PlateOne,
     &                             LengthOfData)
      CALL Dispatch_7(TaskTwo,SubS1,DataAreaOne,BufferLength,PlateOne,
     &                              DataAreaTwo,PlateTwo,
     &                              SourceOne,
     &                              S1)
      CALL Dispatch_7(TaskThree,SubS2,DataAreaOne,BufferLength,PlateOne,
     &                                DataAreaThree,PlateThree,
     &                                SourceTwo,
     &                                S2)
      CALL Dispatch_7(TaskFour,SubS3,DataAreaOne,BufferLength,PlateOne,
     &                               DataAreaFour,PlateFour,
     &                               SourceThree,
     &                               S3)
      CALL Dispatch_8(TaskFive,SubS4,DataAreaTwo,PlateTwo,
     &                               DataAreaThree,PlateThree,
     &                               DataAreaFour,PlateFour,
     &                               SourceFour,
     &                               S4)
C
C  wait for completion of tasks
C
      DO J4TEMP=1,5
         CALL WaitForAnyTask(K4TEMP)
         CALL TerminateTask(K4TEMP)
      END DO
C
C  delete data plates
C
      CALL DeleteDataPlate(PlateOne)
      CALL DeleteDataPlate(PlateTwo)
      CALL DeleteDataPlate(PlateThree)
      CALL DeleteDataPlate(PlateFour)
C
C  output
C
      WRITE(*,'('' S1= '',E12.6)') S1
      WRITE(*,'('' S2= '',E12.6)') S2
      WRITE(*,'('' S3= '',E12.6)') S3
      WRITE(*,'('' S4= '',E12.6)') S4
C
      STOP
      END
      SUBROUTINE Sub0(DataAreaOne,BufferLength,PlateOne,
     &                LengthOfData)
C
C
C  routine to generate array A by a pseudo-random procedure
C  (A)FORTRAN CALL: CALL Sub0(DataAreaOne,BufferLength,PlateOne,
C                             LengthOfData)
C     1.DataAreaOne: <R4> area one for data interchange,
C                         dimension(LengthOfBuffer,2)
C     2.BufferLength: <I4> buffer length
C     3.PlateOne: <I4> data plate carrying the information to
C                      access data in area one
C     4.LengthOfData: <I4> length of array A
C
C  dummy arguments
C
      INTEGER*4 BufferLength,PlateOne,LengthOfData
      REAL*4 DataAreaOne(BufferLength,2)
C
C  private data
C
      INTEGER*4 J4TEMP$,K4TEMP$,Argu_1$,Argu_2$,Argu_3$,Seed$
      AUTOMATIC J4TEMP$,K4TEMP$,Argu_1$,Argu_2$,Argu_3$,Seed$
C
C  find a space to write data
C
      Seed$=666666
      Argu_1$=1
      DO J4TEMP$=1,LengthOfData,BufferLength
         SELECT CASE (Argu_2$)
         CASE (1)
            Argu_2$=2
         CASE DEFAULT
            Argu_2$=1
         END SELECT
C
C  generate data
C
         Argu_3$=MIN0(J4TEMP$+BufferLength-1,LengthOfData)-J4TEMP$+1
         DO K4TEMP$=1,Argu_3$
            DataAreaOne(K4TEMP$,Argu_2$)=Ran(Seed$)
         END DO
C
C  post a data plate
C
         CALL PostDataPlate(PlateOne,Argu_1$,Argu_2$,Argu_3$)
      END DO
C
C  post message of completion
C
      Argu_1$=2
      CALL PostDataPlate(PlateOne,Argu_1$,Argu_2$,Argu_3$)
C
      RETURN
      END
      SUBROUTINE SubS1(DataAreaOne,BufferLength,PlateOne,
     &                 DataAreaTwo,PlateTwo,
     &                 Source,
     &                 S1)
C
C
C  routine to compute S1, and to forward the result to Task_4 when
C  computation is complete
C  (A)FORTRAN CALL: CALL SubS1(DataAreaOne,BufferLength,PlateOne,
C                              DataPlateTwo,PlateTwo,
C                              Source,
C                              S1)
C     1.DataAreaOne: <R4> area one for data interchange,
C                         dimension(BufferLength,2)
C     2.BufferLength: <I4> buffer length of data area one
C     3.PlateOne: <I4> data plate carrying the information to access
C                      data in area one
C     4.DataAreaTwo: <R4> variable for data interchange
C     5.PlateTwo: <I4> data plate carrying the information to access
C                      data in area two
C     6.Source: <I4> list of data plates that will sent to the task,
C                    dimension(*). The list is terminated with a null
C                    value.
C     7.S1: <R4> result
C
C  dummy arguments
C
      INTEGER*4 BufferLength,PlateOne,PlateTwo,Source(1)
      REAL*4 DataAreaOne(BufferLength,2),DataAreaTwo,S1
C
C  private variables
C
      INTEGER*4 ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
      AUTOMATIC ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
C
C  initialize
C
      S1=0.0
C
C  read data plates
C
10    CALL ReadDataPlate(Source,ID$,Argu_1$,Argu_2$,Argu_3$)
C
C  compute S1
C
      IF(ID$.EQ.PlateOne) THEN
         SELECT CASE (Argu_1$)
         CASE (1)
            DO J4TEMP$=1,Argu_3$
               S1=S1+DataAreaOne(J4TEMP$,Argu_2$)
            END DO
C
C  send result and forward message to task 4
C
         CASE (2)
            DataAreaTwo=S1
            CALL PostDataPlate(PlateTwo,INT4(1),J4TEMP$,J4TEMP$)
            CALL PostDataPlate(PlateTwo,INT4(2),J4TEMP$,J4TEMP$)
            CALL FinishDataPlate(ID$)
            RETURN
         END SELECT
      END IF
C
      CALL FinishDataPlate(ID$)
      GO TO 10
      END
      SUBROUTINE SubS2(DataAreaOne,BufferLength,PlateOne,
     &                 DataAreaThree,PlateThree,
     &                 Source,
     &                 S2)
C
C
C  routine to compute S2, and forward the result to Task_4 when
C  computation is complete
C  (A)FORTRAN CALL: CALL SubS2(DataAreaOne,BufferLength,PlateOne,
C                              DataPlateThree,PlateThree,
C                              Source,
C                              S2)
C     1.DataAreaOne: <R4> area one for data interchange,
C                         dimension(BufferLength,2)
C     2.BufferLength: <I4> buffer length of data area one
C     3.PlateOne: <I4> data plate carrying the information to access
C                      data in area one
C     4.DataAreaThree: <R4> variable for data interchange
C     5.PlateThree: <I4> data plate carrying the information to access
C                        data in area three
C     6.Source: <I4> list of data plates that will sent to the task,
C                    dimension(*). The list is terminated with a null
C                    value.
C     7.S2: <R4> result
C
C  dummy arguments
C
      INTEGER*4 BufferLength,PlateOne,PlateThree,Source(1)
      REAL*4 DataAreaOne(BufferLength,2),DataAreaThree,S2
C
C  private variables
C
      INTEGER*4 ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
      AUTOMATIC ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
C
C  initialize
C
      S2=0.0
C
C  read data plates
C
10    CALL ReadDataPlate(Source,ID$,Argu_1$,Argu_2$,Argu_3$)
C
C  compute S2
C
      IF(ID$.EQ.PlateOne) THEN
         SELECT CASE (Argu_1$)
         CASE (1)
            DO J4TEMP$=1,Argu_3$
               S2=S2+DataAreaOne(J4TEMP$,Argu_2$)*
     &               DataAreaOne(J4TEMP$,Argu_2$)
            END DO
C
C  send result and forward message to task 4
C
         CASE (2)
            DataAreaThree=S2
            CALL PostDataPlate(PlateThree,INT4(1),J4TEMP$,J4TEMP$)
            CALL FinishDataPlate(ID$)
            RETURN
         END SELECT
      END IF
C
      CALL FinishDataPlate(ID$)
      GO TO 10
      END
      SUBROUTINE SubS3(DataAreaOne,BufferLength,PlateOne,
     &                 DataAreaFour,PlateFour,
     &                 Source,
     &                 S3)
C
C
C  routine to compute S3, and forward the result to Task_4 when
C  computation is complete
C  (A)FORTRAN CALL: CALL SubS3(DataAreaOne,BufferLength,PlateOne,
C                              DataPlateFour,PlateFour,
C                              Source,
C                              S3)
C     1.DataAreaOne: <R4> area one for data interchange,
C                         dimension(BufferLength,2)
C     2.BufferLength: <I4> buffer length of data area one
C     3.PlateOne: <I4> data plate carrying the information to access
C                      data in area one
C     4.DataAreaFour: <R4> variable for data interchange
C     5.PlateFour: <I4> data plate carrying the information to access
C                       data in area four
C     6.Source: <I4> list of data plates that will sent to the task,
C                    dimension(*). The list is terminated with a null
C                    value.
C     7.S3: <R4> result
C
C  dummy arguments
C
      INTEGER*4 BufferLength,PlateOne,PlateFour,Source(1)
      REAL*4 DataAreaOne(BufferLength,2),DataAreaFour,S3
C
C  private variables
C
      INTEGER*4 ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
      AUTOMATIC ID$,Argu_1$,Argu_2$,Argu_3$,J4TEMP$
C
C  initialize
C
      S3=0.0
C
C  read data plates
C
10    CALL ReadDataPlate(Source,ID$,Argu_1$,Argu_2$,Argu_3$)
C
C  compute S3
C
      IF(ID$.EQ.PlateOne) THEN
         SELECT CASE (Argu_1$)
         CASE (1)
            DO J4TEMP$=1,Argu_3$
               S3=S3+DataAreaOne(J4TEMP$,Argu_2$)*
     &               DataAreaOne(J4TEMP$,Argu_2$)*
     &               DataAreaOne(J4TEMP$,Argu_2$)
            END DO
C
C  send result and forward message to task 4
C
         CASE (2)
            DataAreaFour=S3
            CALL PostDataPlate(PlateFour,INT4(1),J4TEMP$,J4TEMP$)
            CALL FinishDataPlate(ID$)
            RETURN
         END SELECT
      END IF
C
      CALL FinishDataPlate(ID$)
      GO TO 10
      END
      SUBROUTINE SubS4(DataAreaTwo,PlateTwo,
     &                 DataAreaThree,PlateThree,
     &                 DataAreaFour,PlateFour,
     &                 Source,
     &                 S4)
C
C
C  routine to compute S4
C  (A)FORTRAN CALL: CALL SubS4(DataAreaTwo,PlatTwo,
C                              DataPlateThree,PlateThree,
C                              DataPlateFour,PlateFour,
C                              Source,
C                              S4)
C     1.DataAreaTwo: <R4> variable for data interchange
C     2.PlateTwo: <I4> data plate carrying the information to access
C                      data in area two
C     3.DataAreaThree: <R4> variable for data interchange
C     4.PlateThree: <I4> data plate carrying the information to access
C                        data in area three
C     5.DataAreaFour: <R4> variable for data interchange
C     6.PlateFour: <I4> data plate carrying the information to access
C                       data in area four
C     7.Source: <I4> list of data plates that will sent to the task,
C                    dimension(*). The list is terminated with a null
C                    value.
C     8.S4: <R4> result
C
C  dummy arguments
C
      INTEGER*4 PlateTwo,PlateThree,PlateFour,Source(1)
      REAL*4 DataAreaTwo,DataAreaThree,DataAreaFour,S4
C
C  private variables
C
      INTEGER*4 ID$,Argu_1$,Argu_2$,Argu_3$,Coming$
      Logical*4 Done$
      REAL*4 S1$,S2$,S3$
      AUTOMATIC ID$,Argu_1$,Argu_2$,Argu_3$,Done$,Coming$
      AUTOMATIC S1$,S2$,S3$
C
C  initialize
C
      S4=0.0
      Coming$=0
      Done$=.False.
C
C  read data plates
C
10    CALL ReadDataPlate(Source,ID$,Argu_1$,Argu_2$,Argu_3$)
C
C  read data on plate two
C
      IF(ID$.EQ.PlateTwo) THEN
         CALL FinishDataPlate(ID$)
         SELECT CASE (Argu_1$)
         CASE (1)
            S1$=DataAreaTwo
            Coming$=Coming$+1
            IF(Coming$.EQ.3) THEN
               S4=S1$*S2$+S3$
               Coming$=4
               IF(Done$) RETURN
            END IF
         CASE (2)
            IF(Coming$.EQ.4) RETURN
            Done$=.TRUE.
         END SELECT
C
C  read data on plate three
C
      ELSE IF(ID$.EQ.PlateThree) THEN
         S2$=DataAreaThree
         Coming$=Coming$+1
         CALL FinishDataPlate(ID$)
         IF(Coming$.EQ.3) THEN
            S4=S1$*S2$+S3$
            Coming$=4
            IF(Done$) RETURN
         END IF
C
C  read data on plate four
C
      ELSE IF(ID$.EQ.PlateFour) THEN
         S3$=DataAreaFour
         Coming$=Coming$+1
         CALL FinishDataPlate(ID$)
         IF(Coming$.EQ.3) THEN
            S4=S1$*S2$+S3$
            Coming$=4
            IF(Done$) RETURN
         END IF
      END IF
C
      GO TO 10
      END

      REAL*4 FUNCTION RAN(SEED)
C
C
C  function to generate a pseudo-random number
C  (A)FORTRAN ENTRY: RAN(SEED)
C     1.RAN: <R4> return a random number
C     2.SEED: <I4> a seed (updated)
C
      INTEGER*4 SEED,A
      REAL*8 RTEMPA,RTEMPB
      DATA M/2147483647/,A/16807/
C
C  generate a random number
C
      RTEMPA=DFLOAT(SEED)*A
      ITEMP=INT(RTEMPA/DFLOAT(M))
      RTEMPB=DFLOAT(M)*DFLOAT(ITEMP)
      SEED=INT(RTEMPA-RTEMPB)
      RAN=DFLOAT(SEED)/DFLOAT(M)
C
      RETURN
      END
 
